/*
 * calccore/scc/basefunc.cc
 * 
 * Simple calc core baseic object/value functionality. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "ccif_simple.h"

#include <core/execthread.h>

#include <numerics/spline/cspline.h>

#include <objdesc/omatrix.h>

#include <hlib/refstrhash.h>


namespace CC { namespace S
{

static const char *tid2str[TID_LAST]=
{
	"[none]",
	
	"accel_scale",
	"addpts",
	"constant",
	"cspline",
	"curve",
	"dt",
	"fixed",
	"force_center",
	"force_up",
	"front",
	"front_mode",
	"function",
	"gravity",
	"interpol",
	"interpol_npts",
	"join",
	"length",
	"lspline",
	"none",
	"pos",
	"pts",
	"pts_t",
	"speed",
	"t",
	"tvals",
	"up",
	"up_mode",
};

// Hash for quick lookup and reference counter. 
static RefStringHash<AniDescBlockTokID> *tid2str_hash=NULL;
static int tid2str_hash_refs=0;
static size_t tid2str_hash_lookups=0;


const char *AniDescBlockTokIDString(AniDescBlockTokID tid)
{
	assert(tid>=0 && tid<TID_LAST);
	return(tid2str[tid]);
}

AniDescBlockTokID LookupAniDescBlockTokID(const RefString &str)
{
	if(tid2str_hash)
	{
		++tid2str_hash_lookups;
		AniDescBlockTokID tok_id;
		if(!tid2str_hash->lookup(str,&tok_id,/*lru_requeue=*/0))
		{  return(tok_id);  }
	}
	else
	{
		assert(0);  // We want to use the hash. Non-hash lookup works, too. 
		for(int i=0; i<TID_LAST; i++)
		{
			if(str==tid2str[i])
			{  return((AniDescBlockTokID)i);  }
		}
	}
	return(TID_None);
}

//------------------------------------------------------------------------------


CCBaseFunctionality::InterpreteScopeEntries_IData::InterpreteScopeEntries_IData(
	ANI::TreeNode *first)
{
	curr_tn=first;
	curr_ent_idx=-1;
	next_ent_idx=0;
	curr_ent_cat=-1;
	next_ent_cat=0;
}


int CCBaseFunctionality::InterpreteScopeEntries(
	ADTNTypedScope *tsc,ExecThread *et,
	const char *scope_name,const ScopeEntryCategory *cats,
	void *dptr,int slot_id)
{
	InterpreteScopeEntries_IData *idata=
		(InterpreteScopeEntries_IData*)et->ess->popPtr(NULL);
	if(!idata)
	{  idata=new InterpreteScopeEntries_IData(tsc->first_entry);  }
	
	for(ANI::TreeNode *_tn=idata->curr_tn; _tn; _tn=_tn->next)
	{
		if(_tn->NType()!=ANI::TN_AniDesc) assert(0);
		if(((ADTreeNode*)_tn)->ADNType()!=ADTN_ScopeEntry) assert(0);
		
		ADTNScopeEntry *scent=(ADTNScopeEntry*)_tn;
		ANI::TNIdentifier *name = (scent->down.first()==scent->down.last() ? 
			NULL : (ANI::TNIdentifier*)scent->down.first());
		ANI::TreeNode *value=scent->down.last();
		
		int ent_idx=idata->curr_ent_idx;
		int ent_cat=idata->curr_ent_cat;
		if(ent_idx<0)  // Need to look up. 
		{
			RefString name_str;
			if(name) name_str=name->CompleteStr();
			
			if(name_str)
			{
				// NOTE: This should be done at parsing time and not at interpretation 
				//       time! Or here in a cache-like manner. 
				#warning FIXME (noncrtical). 
				AniDescBlockTokID tok_id=LookupAniDescBlockTokID(name_str);
				
				// Look up the token (seee if it is supported and in which 
				// category it is): 
				// (Note the same token can be in different categories 
				// for different scopes.) 
				if(tok_id!=TID_None) for(int j=0; cats[j].toks; j++)
				{
					const ScopeEntryCategory *c=&cats[j];
					for(int i=0; c->toks[i]!=TID_None; i++)
						if(c->toks[i]==tok_id)
						{  ent_cat=j;  ent_idx=i;  goto breakout;  }
				}
				breakout:;
			}
			else if(idata->next_ent_idx>=0)
			{
				// No identifer specified. Assume next one if available. 
				ent_idx=idata->next_ent_idx;
				ent_cat=idata->next_ent_cat;
			}
			
			if(ent_idx<0)
			{
				_UnknownProperty(name,value,scope_name);  // increases error counter
				
				idata->next_ent_idx=-1;
				idata->next_ent_cat=-1;
				//idata->curr_ent_idx=-1;   // <-- already the case
				continue;
			}
			
			idata->next_ent_idx=ent_idx+1;
			idata->next_ent_cat=ent_cat;  // NOT +1
			// Don't run out of range: 
			assert(ent_cat>=0);
			if(cats[ent_cat].toks[idata->next_ent_idx]==TID_None)
			{
				idata->next_ent_idx=0;
				++idata->next_ent_cat;
				if(!cats[idata->next_ent_cat].toks)
				{  idata->next_ent_idx=-1;  idata->next_ent_cat=-1;  }
			}
		}
		assert(ent_idx>=0 && ent_cat>=0);
		
		
		// Okay, we cannot do more here, need to go 
		// back to the calling level. 
		InterpreteScopeEntriesInfo info;
		info.cat=&cats[ent_cat];
		info.tok_id=cats[ent_cat].toks[ent_idx];
		info.value=value;
		info.tsc=NULL;
		info.tsc_tok_id=TID_None;
		// We may need the type of the value part: 
		if(value)
		{
			if(value->NType()==ANI::TN_AniDesc && 
				((ADTreeNode*)value)->ADNType()==ADTN_TypedScope)
			{
				info.tsc=(ADTNTypedScope*)value;
				info.tsc_tok_id=LookupAniDescBlockTokID(
					info.tsc->scope_type->name);
			}
		}
		info.dptr=dptr;
		InterpreteScopeSlot_State rv=(InterpreteScopeSlot_State)
			InterpreteScopeSlot(slot_id,et,&info);
		switch(rv)
		{
			case ISS_Cont:  break;
			case ISS_OOPSError:
				// This may not happen. 
				assert(0);
				++n_errors;
				break;
			case ISS_CSwitch:
				// Context switch. 
				idata->curr_ent_idx=ent_idx;
				idata->curr_ent_cat=ent_cat;
				idata->curr_tn=_tn;
				et->ess->pushPtr(idata);
				return(1);
			case ISS_AlreadySet:
				Error(value->GetLocationRange(),
					"%s scope property \"%s\" was already set\n",
					scope_name,
					AniDescBlockTokIDString(cats[ent_cat].toks[ent_idx]));
				++n_errors;
				break;
			case ISS_TimeSpecConflict:
				Error(value->GetLocationRange(),
					"conflicting %s time specs\n",
					scope_name);
				++n_errors;
				break;
			case ISS_IllegalValue:
				Error(value->GetLocationRange(),
					"illegal value for \"%s\" property in %s scope\n",
					AniDescBlockTokIDString(cats[ent_cat].toks[ent_idx]),
					scope_name);
				++n_errors;
				break;
			default:  assert(0);
		}
		
		// Done this entry: 
		idata->curr_ent_idx=-1;
		idata->curr_ent_cat=-1;
	}
	
	delete idata;
	return(0);
}


int CCBaseFunctionality::InterpreteScopeSlot(
		int /*slot_id*/,ExecThread * /*et*/,InterpreteScopeEntriesInfo * /*info*/)
{
	// Must be overridden when used. 
	assert(0);
	return(ISS_OOPSError);
}


void CCBaseFunctionality::_UnknownProperty(ANI::TNIdentifier *name,
	ANI::TreeNode *value,const char *scope_name)
{
	if(name)
	{
		Error(name->GetLocationRange(),
			"unknown property \"%s\" in %s scope [ignored]\n",
			name->CompleteStr().str(),scope_name);
	}
	else
	{
		Error(value->GetLocationRange(),
			"anonymous property in %s scope [ignored]\n",
			scope_name);
	}
	++n_errors;
}


// Eval scalar expression. 
// Return value: 
//   0 -> OK, value read and stored in *ret_val
//   1 -> context switch
//  -1 -> error: message written, n_errors incremented and *ret_val 
//        not modified
int CCBaseFunctionality::_GetScalarValue(double &ret_val,ANI::TreeNode *tn,
	ExecThread *et)
{
	do {
		if(tn->NType()!=ANI::TN_Expression)  break;
		
		ANI::TNExpression *expr=(ANI::TNExpression*)tn;
		ANI::ExprValue tmp;
		if(expr->Eval(tmp,et))
		{  return(1);  }
		
		if(tmp.GetExprValueTypeID()!=ANI::EVT_POD)  break;
		
		Value valtmp;
		tmp.GetPOD(&valtmp);
		if(valtmp.Type()!=Value::VTScalar && 
		   valtmp.Type()!=Value::VTInteger)
		{  break;  }
		
		ret_val=valtmp.GetScalar().val();  // Will convert int->double if needed. 
		return(0);
	} while(0);
	
	//tn->DumpTree(stdout);
	Error(tn->GetLocationRange(),
		"invalid argument: (scalar) expression expected\n");
	++n_errors;
	return(-1);
}


// Like _GetScalarValue for strings. 
int CCBaseFunctionality::_GetStringValue(String &ret_val,ANI::TreeNode *tn,
	ExecThread *et)
{
	do {
		if(tn->NType()!=ANI::TN_Expression)  break;
		
		ANI::TNExpression *expr=(ANI::TNExpression*)tn;
		ANI::ExprValue tmp;
		if(expr->Eval(tmp,et))
		{  return(1);  }
		
		if(tmp.GetExprValueTypeID()!=ANI::EVT_POD)  break;
		
		Value valtmp;
		tmp.GetPOD(&valtmp);
		if(valtmp.Type()!=Value::VTString)
		{  break;  }
		
		ret_val=valtmp.GetString();
		return(0);
	} while(0);
	
	//tn->DumpTree(stdout);
	Error(tn->GetLocationRange(),
		"invalid argument: (string) expression expected\n");
	++n_errors;
	return(-1);
}


// Like _GetScalarValue for range. 
int CCBaseFunctionality::_GetRangeValue(Range &ret_val,ANI::TreeNode *tn,
	ExecThread *et)
{
	do {
		if(tn->NType()!=ANI::TN_Expression)  break;
		
		ANI::TNExpression *expr=(ANI::TNExpression*)tn;
		ANI::ExprValue tmp;
		if(expr->Eval(tmp,et))
		{  return(1);  }
		
		if(tmp.GetExprValueTypeID()!=ANI::EVT_POD)  break;
		
		Value valtmp;
		tmp.GetPOD(&valtmp);
		if(valtmp.Type()!=Value::VTRange)
		{  break;  }
		
		ret_val=valtmp.GetRange();
		return(0);
	} while(0);
	
	//tn->DumpTree(stdout);
	Error(tn->GetLocationRange(),
		"invalid argument: (range) expression expected\n");
	++n_errors;
	return(-1);
}


// Like _GetScalarValue for vector<3>. 
int CCBaseFunctionality::_GetVector3Value(double *vector,ANI::TreeNode *tn,
	ExecThread *et)
{
	do {
		if(tn->NType()!=ANI::TN_Expression)  break;
		
		ANI::TNExpression *expr=(ANI::TNExpression*)tn;
		ANI::ExprValue tmp;
		if(expr->Eval(tmp,et))
		{  return(1);  }
		
		if(tmp.GetExprValueTypeID()!=ANI::EVT_POD)  break;
		
		Value valtmp;
		tmp.GetPOD(&valtmp);
		if(valtmp.Type()==Value::VTScalar || 
		   valtmp.Type()==Value::VTInteger)
		{
			// Will convert int->double if needed. 
			double tmp=valtmp.GetScalar().val();
			// Promote vector: 
			for(int i=0; i<3; i++)
				vector[i]=tmp;
			return(0);
		}
		if(valtmp.Type()==Value::VTVector)
		{
			const Vector *vect=valtmp.GetPtr<Vector>();
			assert(vect->dim()==3);
			for(int i=0; i<3; i++)
				vector[i]=(*vect)[i];
			return(0);
		}
	} while(0);
	
	//tn->DumpTree(stdout);
	Error(tn->GetLocationRange(),
		"invalid argument: (vector<3>) expression expected\n");
	++n_errors;
	return(-1);
}


// Eval vector array expression. 
// Return value: see _GetScalarValue(). 
int CCBaseFunctionality::_GetVectorArray(NUM::VectorArray<double> &array,
	ANI::TreeNode *tn,ExecThread *et,int vec_dim)
{
	do {
		if(tn->NType()!=ANI::TN_Expression)  break;
		
		ANI::TNExpression *expr=(ANI::TNExpression*)tn;
		ANI::ExprValue tmp;
		if(expr->Eval(tmp,et))
		{  return(1);  }
		
		ANI::ExprValueType evt;
		tmp.GetExprValueType(&evt);
		//fprintf(stderr,"TYPE=%s\n",evt.TypeString().str());
		const Value::CompleteType *pod_elem_type=evt.ArrayType().PODType();
		if(!pod_elem_type || pod_elem_type->type!=Value::VTVector)  break;
		
		ANI::ExprValueArray tmparr;
		tmp.GetArray(&tmparr);
		ssize_t arrsize=tmparr.GetSize();
		assert(arrsize>=0);
		
		const int vdim=pod_elem_type->n;
		assert(vdim>=1);
		if(vec_dim>=0 && vdim!=vec_dim)  break;
		array.Resize(arrsize,vdim);
	
		ANI::ExprValue elemtmp;
		Value vecttmp;
		for(ssize_t idx=0; idx<arrsize; idx++)
		{
			int rv=tmparr.IndexGet(idx,&elemtmp);
			assert(!rv);
			elemtmp.GetPOD(&vecttmp);
			const Vector *vect=vecttmp.GetPtr<Vector>();
			assert(vect->dim()==vdim);
			for(int i=0; i<vdim; i++)
			{  array[idx][i]=(*vect)[i];  }
		}
		
		return(0);
	} while(0);
	
	//tn->DumpTree(stdout);
	char tmpstr[24];
	if(vec_dim<0)
	{  *tmpstr='\0';  }
	else
	{  snprintf(tmpstr,24,"<%d>",vec_dim);  }
	Error(tn->GetLocationRange(),
		"invalid argument: (vector%s[]) expression expected\n",tmpstr);
	++n_errors;
	return(-1);
}


int CCBaseFunctionality::AquireADBLock(ADTNAniDescBlock *adb,ExecThread *et)
{
	if(this->lock_thread!=et->thread_id || this->lock_adb!=adb)
	{
		// The object is locked by another thread. 
		// First, see if this other thread still exists or if it 
		// died in the meantime. 
		#warning FIXME FIXME FIXME FIXME FIXME
		// FIXME: Need better locking mechanism. thread id may be re-used 
		//        and is only unique on a per-context basis. 
		// NOTE: Objects can be shared between different contexts (via 
		//       static pointers). 
		// Problem: Threads suspended in ADBs during context switch, e.g. 
		// between two settings. 
		// Und was ist berhaupt mit dem CCObject, wenn es von mehreren 
		// contexts bearbeitet wird...? Scheie. 
		// N... keine Scheie: der ADB bleibt eben dann gelockt, bis der 
		// ursprngliche context wieder dran ist. 
		
		if(!et->TestThreadExist(this->lock_thread))
		{  LoseADBLock();  }
		else
		{
			Error(adb->GetLocationRange(),
				"ADB cannot be executed by thread [%d] because...\n",
				this->lock_thread);
			Error(this->lock_adb->GetLocationRange(),
				"...thread [%d] currently holds ADB lock [skipped]\n",
				this->lock_thread);
			return(1);
		}
	}
	
	if(this->lock_thread<0)
	{
		assert(!lock_adb);
		this->lock_thread=et->thread_id;
		this->lock_adb=adb;
		assert(et->thread_id>0);  // thread 0 is non-ani context
		n_errors=0;
	}
	
	return(0);
}


void CCBaseFunctionality::LoseADBLock()
{
	this->lock_thread=-1;
	this->lock_adb=NULL;
	// n_errors not yet cleared. 
}


CCBaseFunctionality::CCBaseFunctionality()
{
	lock_thread=-1;
	lock_adb=NULL;
	n_errors=0;

	// Init token lookup hash: 
	if(tid2str_hash)
	{  ++tid2str_hash_refs;  }
	else
	{
		fprintf(stdout,"SCC: Setting up identifier hash: ");
		
		tid2str_hash=new RefStringHash<AniDescBlockTokID>(
			/*hash_size=*/101);
		tid2str_hash_refs=1;
		
		// Note: Beginning loop at 1 is correct; we don't need "[none]"
		//       in the hash. 
		for(int i=1; i<TID_LAST; i++)
		{
			RefString tmp;
			tmp.set(tid2str[i]);
			tid2str_hash->store(tmp,(AniDescBlockTokID)i,
					/*lru_requeue=*/0,/*dont_search=*/1);
		}

		fprintf(stdout,"%d entries\n",tid2str_hash->count());
	}
}

CCBaseFunctionality::~CCBaseFunctionality()
{
	//assert(lock_thread<0 && !lock_adb); <-- NO; thread may have been killed

	// Clean up hash: 
	if(!--tid2str_hash_refs)
	{
		fprintf(stdout,"SCC: Cleaning up identifier hash (%u lookups).\n",
			tid2str_hash_lookups);
		DELETE(tid2str_hash);
	}
}

}}  // end of namespace CC::S
