/*
 * calccore/ccif_simple.h
 * 
 * Simple calc core (internal) setting/object/value description. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _ANIVISION_CALCCORE_CCS_CCIF_H_
#define _ANIVISION_CALCCORE_CCS_CCIF_H_ 1

#include <hlib/linkedlist.h>
#include <calccore/ccif.h>

#include <core/execthread.h>

#include <calccore/adtree.h>
//#include <numerics/la_basics.h>
#include <numerics/spline/sl_base.h>

#include <objdesc/move.h>


namespace CC
{
namespace S
{

// As defined in this file: 
class CCSetting;
class CCObject;
class CCValue;


// All the token IDs recognized: 
enum AniDescBlockTokID
{
	// MAKE SURE THESE ARE INDEXED 0..TID_LAST. 
	TID_None=0,
	
	TID_accel_scale,
	TID_addpts,
	TID_constant,
	TID_cspline,
	TID_curve,
	TID_dt,
	TID_fixed,
	TID_force_center,
	TID_force_up,
	TID_front,
	TID_front_mode,
	TID_function,
	TID_gravity,
	TID_interpol,
	TID_interpol_npts,
	TID_join,
	TID_length,
	TID_lspline,
	TID_none,
	TID_pos,
	TID_pts,
	TID_pts_t,
	TID_speed,
	TID_t,
	TID_tvals,
	TID_up,
	TID_up_mode,
	
	TID_LAST
};

// Translate AniDescBlockTokID into string representation 
// and look up string to get AniDescBlockTokID (or TID_None if 
// not found): 
extern const char *AniDescBlockTokIDString(AniDescBlockTokID tid);
// Lokup ani desc tokens: This is actually using a hash to 
// speed things up and returns TID_None if no match. 
extern AniDescBlockTokID LookupAniDescBlockTokID(const RefString &str);


// The different (token) categories (cat_id): 
// The *Base categories are also used as slot IDs. 
enum AniDescBlockCatID
{
	CID_MoveBase,              // --> toks_move_base
	CID_CurveBase,             // --> toks_curve_base
	CID_CurvePosFixedBase,     // --> toks_curve_pos_fixed_base
	CID_CurvePosLsplineBase,   // --> toks_curve_pos_lspline_base
	CID_CurvePosCsplineBase,   // --> toks_curve_pos_cspline_base
	CID_CurveTMapNoneBase,     // --> curve_tmap_none_cats
	CID_CurveTMapConstBase,    // --> curve_tmap_const_cats
	CID_FrontSpeedBase,        // --> front_speed_cats
	CID_UpGravityBase,         // --> up_gravity_cats
};


struct CCBaseFunctionality
{
	public:
		struct ScopeEntryCategory
		{
			// To recognize the category: 
			AniDescBlockCatID cat_id;
			
			// Array storing all allowed (recognized) tokens 
			// in their default order. 
			// Array must be terminated by TID_None. 
			const AniDescBlockTokID *toks;
		};
		
		// Internally used by NextScopeEntry(): 
		struct InterpreteScopeEntries_IData
		{
			ANI::TreeNode *curr_tn;  // where we are in the loop
			int curr_ent_cat;   // Category index, not ID. 
			int curr_ent_idx;
			int next_ent_cat;   // Category index, not ID. 
			int next_ent_idx;
			
			_CPP_OPERATORS
			InterpreteScopeEntries_IData(ANI::TreeNode *first);
			~InterpreteScopeEntries_IData() {}
		};
		
		struct InterpreteScopeEntriesInfo
		{
			const ScopeEntryCategory *cat;   // ...of the token (entry) below
			AniDescBlockTokID tok_id;   // found entry (token)
			ANI::TreeNode *value;       // value part
			void *dptr;                 // as passed to InterpreteScopeEntries()
			// In case value is an ADTNTypedScope, the following is also set: 
			// (Otherwise tsc is NULL) 
			ADTNTypedScope *tsc;   // =value, casted
			AniDescBlockTokID tsc_tok_id;  // ID of tsc->scope_type->name
		};
		
		enum InterpreteScopeSlot_State
		{
			// DO NOT CHANGE the sign of these constants: 
			ISS_OOPSError=-1,   // uncaught error; may not happen...
			ISS_Cont=0,
			ISS_CSwitch=1,
			// Error codes with special error message being written: 
			ISS_AlreadySet=-2,  // report "property was already set" error
			ISS_TimeSpecConflict=-3,  // report "conflicting time specs" error
			ISS_IllegalValue=-4,  // reort "illegal value for ..." error
		};
	protected:
		// thread_id of the thread currently executing an ADB for 
		// this object. -1 for "none". 
		int lock_thread;
		// Pointer to the ADB currently being executed. 
		ADTNAniDescBlock *lock_adb;
		// Error counter for current ADB execution (see lock above). 
		int n_errors;
		
		// We use the EvalStateStack to temporarily store curve 
		// objects across context switches. Cloning makes no 
		// trouble as it is forbidden inside ADBs anyways and 
		// the pushPtr()/popPtr() does not allow cloning at all. 
		// These are helper functions: 
		inline void _PushCPtr(ExecThread *et,void *ptr)
			{  et->ess->pushPtr(ptr);  }
		template<class T>inline T *_PopCPtr(ExecThread *et)
			{  T *ptr=(T*)et->ess->popPtr(NULL); return(ptr ? ptr : new T()); }
		
		// Lock/unlock ADB. Lock function returns 1 if an error 
		// happened and the lock could not be obtained. 
		// AquireADBLock() will also clear n_errors counter. 
		int AquireADBLock(ADTNAniDescBlock *adb,ExecThread *et);
		void LoseADBLock();  // <-- Does NOT clear n_errors. 
		
		// Helper for scope entry parsing. Get next entry. 
		// scope_name: The name of the scope we're currently in ("curve",...)
		//             Only needed for error messages. 
		// cats: Array of ScopeEntryCategories terminated by an element 
		//       with cats.toks=NULL. Each ScopeEntryCategory holds an 
		//       array of AniDescBlockTokIDs terminated by an element with 
		//       a TID_None token ID. The AniDescBlockTokIDs describe the 
		//       possible scope entries (properties, tokens). 
		//   Note that order of the entries is important because in case 
		//   an unnamed entry is found, it is considered to be the successor 
		//   of the previous one. 
		// For each entry, the virtual function InterpreteScopeSlot() 
		// is called with slot_id as first argument and dptr in the 
		// info parameter (see there). 
		// Return value: 
		//    0 -> OK, done
		//    1 -> context switch
		int InterpreteScopeEntries(
			ADTNTypedScope *tsc,ExecThread *et,
			const char *scope_name,const ScopeEntryCategory *cats,
			void *dptr,int slot_id);
		// Used by InterpreteScopeEntries(): 
		// Return value is actually an enum InterpreteScopeSlot_State. 
		//    0 -> OK
		//    1 -> context switch
		//  ...more...
		virtual int InterpreteScopeSlot(int slot_id,ExecThread *et,
			InterpreteScopeEntriesInfo *info);
	public:  _CPP_OPERATORS
		CCBaseFunctionality();
		virtual ~CCBaseFunctionality();
		
		// Eval scalar/vector array/string/... expression. 
		// For vector array, vec_dim is the desired vector space dimension 
		// or -1 for "any". 
		// Return value: 
		//   0 -> OK, value read and stored in ret_val
		//   1 -> context switch
		//  -1 -> error: message written, n_errors incremented and ret_val 
		//        not modified
		int _GetScalarValue(double &ret_val,ANI::TreeNode *tn,ExecThread *et);
		int _GetStringValue(String &ret_val,ANI::TreeNode *tn,ExecThread *et);
		int _GetRangeValue(Range &ret_val,ANI::TreeNode *tn,ExecThread *et);
		int _GetVector3Value(double *vector,ANI::TreeNode *tn,ExecThread *et);
		int _GetVectorArray(NUM::VectorArray<double> &array,
			ANI::TreeNode *tn,ExecThread *et,int vec_dim=-1);
		
		// Write unknown property error and increment error 
		// counter n_errors. name may be NULL. 
		void _UnknownProperty(ANI::TNIdentifier *name,
			ANI::TreeNode *value,const char *scope_name);
};


class CCObject : 
	public CCIF_Object,
	public LinkedListBase<CCObject>,
	public CCBaseFunctionality
{
	private:
		// List of movements: 
		LinkedList<NUM::OMovement> movements;
		
		// Movement cache: the movement looked up previously: 
		// Internally used by _FindMovementByTime(). 
		NUM::OMovement *_prev_looked_up_movement;
		
		// Look up a movement node by passed setting time. 
		// Returns NULL if none matches. 
		NUM::OMovement *_FindMovementByTime(double setting_time);
		
		template<typename T,typename B> inline int _InterpreteScope_Template(
			const char *scope_name,const ScopeEntryCategory *cats,int slot_id,
			ADTNTypedScope *tsc,ExecThread *et,B **_dptr);
		
		int _InterpreteCurvePosFixed(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		int _InterpreteCurvePosFixed_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		int _InterpreteCurvePosLspline(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		int _InterpreteCurvePosCspline(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		int _InterpreteCurvePosLCspline_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		int _InterpreteCurvePosFunction(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		int _InterpreteCurvePosCurve(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		int _InterpreteCurvePosJoin(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveFunctionBase **_curvefunc);
		
		int _InterpreteCurveTMapNone(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveTMapFunction **_tmapfunc);
		int _InterpreteCurveTMapNone_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		int _InterpreteCurveTMapConst(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveTMapFunction **_tmapfunc);
		int _InterpreteCurveTMapConst_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		int _InterpreteCurveTMapCspline(ADTNTypedScope *tsc,ExecThread *et,NUM::CurveTMapFunction **_tmapfunc);
		
		int _InterpreteCurvePos(InterpreteScopeEntriesInfo *info,ExecThread *et,NUM::CurveFunctionBase **_posfunc);
		int _InterpreteCurveTMap(InterpreteScopeEntriesInfo *info,ExecThread *et,NUM::CurveTMapFunction **_tmapfunc,bool is_speed);
		
		int _InterpreteCurve(ADTNTypedScope *tsc,ExecThread *et,NUM::OCurve **_ocurve);
		int _InterpreteCurve_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		
		int _InterpreteFrontSpeed(ADTNTypedScope *tsc,ExecThread *et,NUM::OFrontSpec **_ofront);
		int _InterpreteFrontSpeed_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		
		int _InterpreteUpGravity(ADTNTypedScope *tsc,ExecThread *et,NUM::OUpSpec **_oup);
		int _InterpreteUpGravity_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		
		int _InterpreteMovePos(InterpreteScopeEntriesInfo *info,ExecThread *et,NUM::OCurve **_ocurve);
		int _InterpreteMoveFront(InterpreteScopeEntriesInfo *info,ExecThread *et,NUM::OFrontSpec **_ocurve);
		int _InterpreteMoveUp(InterpreteScopeEntriesInfo *info,ExecThread *et,NUM::OUpSpec **_oup);
		
		int _InterpreteMoveScope(ADTNTypedScope *tsc,ExecThread *et,NUM::OMovement **_movement);
		int _InterpreteMoveScope_Slot(ExecThread *et,InterpreteScopeEntriesInfo *info);
		
		// [overriding virtual]
		virtual int InterpreteScopeSlot(int slot_id,ExecThread *et,
			InterpreteScopeEntriesInfo *info);
		
		// Helper for EvalAniDescBlock(); internal use ONLY: 
		int _DoEvalAniDescBlock(ADTNAniDescBlock *adb,ExecThread *et);
	public:   _CPP_OPERATORS
		CCObject(CCSetting *_setting,int _vflags);
		~CCObject();
		
		// Get associated setting: 
		inline CCSetting *Setting()
			{  return((CCSetting*)setting);  }
		
		// [overriding virtuals from CCIF_Object; see there.]
		int GetVal(OVal ov,int *store_here);
		int GetVal(OVal ov,double *store_here);
		int GetVal(OVal ov,NUM::ObjectMatrix *store_here);
		
		// [overriding virtuals from CCIF_Object; see there.]
		int EvalAniDescBlock(ADTNAniDescBlock *adb,ExecThread *et);
};


class CCValue : 
	public CCIF_Value,
	public LinkedListBase<CCValue>,
	public CCBaseFunctionality
{
	private:
		
	public:  _CPP_OPERATORS
		CCValue(CCSetting *_setting,int _dimension);
		~CCValue();
		
		// Get associated setting: 
		inline CCSetting *Setting()
			{  return((CCSetting*)setting);  }
		
		// [overriding virtuals from CCIF_Value; see there.]
};


class CCSetting : 
	public CCIF_Setting
{
	private:
		// Child objects: 
		LinkedList<CCObject> objlist;
		// Child vals: 
		LinkedList<CCValue> vallist;
		
		// Current time range: 
		Range timerange;
		
		// Counts time steps (->TimeStep()) so far: 
		int n_time_steps;
		
	public:  _CPP_OPERATORS
		CCSetting();
		~CCSetting();
		
		// [overriding virtuals from CCIF_Setting; see there.]
		int SetTimeRange(const Range &tr);
		int ComputeAll();
		int TimeStep();
		
		// For setting/value (un)registration (done by their 
		// con/destructor): 
		void RegisterValue(CCValue *val);
		void UnregisterValue(CCValue *val);
		void RegisterObject(CCObject *obj);
		void UnregisterObject(CCObject *obj);
};

}  // end of namespace S

class CCIF_Factory_Simple : public CCIF_Factory
{
	private:
	public:  _CPP_OPERATORS
		CCIF_Factory_Simple();
		~CCIF_Factory_Simple();
		
		// [overriding virtuals from CCIF_Factory:]
		CCIF_Setting *CreateCCIF_Setting();
		CCIF_Object *CreateCCIF_Object(CCIF_Setting *setting,int vflags);
		CCIF_Value *CreateCCIF_Value(CCIF_Setting *setting,int dimension);
};

}  // end of namespace CC

#endif  /* _ANIVISION_CALCCORE_CCS_CCIF_H_ */
