/*
 * calccore/scc/object.cc
 * 
 * Simple calc core object main file. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "ccif_simple.h"

#include <objdesc/omatrix.h>

#include <objdesc/curve_fixed.h>
#include <objdesc/curve_lspline.h>
#include <objdesc/curve_cspline.h>

#include <objdesc/curvetmap_none.h>
#include <objdesc/curvetmap_linear.h>

#include <objdesc/frontspec_speed.h>

#include <objdesc/upspec_gravity.h>


namespace CC { namespace S
{

// Basic curve properties: 
static const AniDescBlockTokID toks_move_base[]=
	{ TID_pos, TID_front, TID_front_mode, TID_up, TID_up_mode,    TID_None };

// Basic curve properties: 
static const AniDescBlockTokID toks_curve_base[]=
	{ TID_pos, TID_dt, TID_t, TID_length, TID_speed,    TID_None };

// Basic (curve pos) fixed properties: 
static const AniDescBlockTokID toks_curve_pos_fixed_base[]=
	{ TID_pos,    TID_None };

// Basic (curve pos) lspline properties: 
static const AniDescBlockTokID toks_curve_pos_lspline_base[]=
	{ TID_pts, TID_pts_t, TID_addpts, TID_tvals,    TID_None };

// And so on...
static const AniDescBlockTokID toks_curve_pos_cspline_base[]=
	{ TID_pts, TID_pts_t, TID_addpts, TID_tvals,    TID_None };

static const AniDescBlockTokID toks_up_gravity_base[]=
	{ TID_force_up, TID_force_center, TID_accel_scale,    TID_None };

// This lists all token recognized inside the move ADB: 
static const CCBaseFunctionality::ScopeEntryCategory move_cats[]=
{
	{ CID_MoveBase, toks_move_base },
	{ (AniDescBlockCatID)-1, NULL }
};

// This lists all token recognized inside the curve block: 
static const CCBaseFunctionality::ScopeEntryCategory curve_cats[]=
{
	{ CID_CurveBase, toks_curve_base },
	{ (AniDescBlockCatID)-1, NULL }
};

// And so on...
static const CCBaseFunctionality::ScopeEntryCategory curve_pos_fixed_cats[]=
{
	{ CID_CurvePosFixedBase, toks_curve_pos_fixed_base },
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory curve_pos_lspline_cats[]=
{
	{ CID_CurvePosLsplineBase, toks_curve_pos_lspline_base },
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory curve_pos_cspline_cats[]=
{
	{ CID_CurvePosCsplineBase, toks_curve_pos_cspline_base },
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory curve_tmap_none_cats[]=
{
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory curve_tmap_const_cats[]=
{
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory front_speed_cats[]=
{
	{ (AniDescBlockCatID)-1, NULL }
};

static const CCBaseFunctionality::ScopeEntryCategory up_gravity_cats[]=
{
	{ CID_UpGravityBase, toks_up_gravity_base },
	{ (AniDescBlockCatID)-1, NULL }
};


template<typename T,typename B> inline int CCObject::_InterpreteScope_Template(
	const char *scope_name,const ScopeEntryCategory *cats,int slot_id,
	ADTNTypedScope *tsc,ExecThread *et,B **_dptr)
{
	// This will create a new one if there is none on the stack. 
	T *dptr = _PopCPtr<T>(et);
	
	int rv=InterpreteScopeEntries(tsc,et, scope_name,cats, dptr,slot_id);
	if(rv>0)  // Context switch. 
	{
		_PushCPtr(et,dptr);
		return(1);
	}
	assert(rv==0);
	
	*_dptr=dptr;
	return(0);
}

// The tsc MUST be a "fixed { ... }" scope, i.e. 
// tsc->scope_type->name must be "fixed". 
int CCObject::_InterpreteCurvePosFixed(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	return(_InterpreteScope_Template<NUM::CurveFunction_Fixed>(
		"(curve pos) fixed",curve_pos_fixed_cats,
		/*slot_id=*/CID_CurvePosFixedBase,
		tsc,et,/*dptr=*/_curvefunc));
}

int CCObject::_InterpreteCurvePosFixed_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	NUM::CurveFunction_Fixed *curvefunc=
		(NUM::CurveFunction_Fixed*)info->dptr;
	
	assert(info->cat->cat_id==CID_CurvePosFixedBase);  // for now until we accept more
	switch(info->tok_id)
	{
		case TID_pos:
		{
			double tmp[3];
			int rv=_GetVector3Value(tmp,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(curvefunc->SetPos(tmp))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		default: assert(0);
	}
	
	return(ISS_Cont);
}


// The tsc MUST be a "lspline { ... }" scope, i.e. 
// tsc->scope_type->name must be "lspline". 
int CCObject::_InterpreteCurvePosLspline(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	return(_InterpreteScope_Template<NUM::CurveFunction_LSpline>(
		"(curve pos) lspline",curve_pos_lspline_cats,
		/*slot_id=*/CID_CurvePosLsplineBase,
		tsc,et,/*dptr=*/_curvefunc));
}

// The tsc MUST be a "cspline { ... }" scope, i.e. 
// tsc->scope_type->name must be "cspline". 
int CCObject::_InterpreteCurvePosCspline(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	return(_InterpreteScope_Template<NUM::CurveFunction_CSpline>(
		"(curve pos) cspline",curve_pos_cspline_cats,
		/*slot_id=*/CID_CurvePosCsplineBase,
		tsc,et,/*dptr=*/_curvefunc));
}

int CCObject::_InterpreteCurvePosLCspline_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	// NOTE: This function processes LINEAR and CUBIC splines. 
	//       See spline_order. 
	
	NUM::CurveFunction_Spline_Base *curvefunc=
		(NUM::CurveFunction_Spline_Base*)info->dptr;
	
	switch(info->cat->cat_id)
	{
		case CID_CurvePosLsplineBase:  // ok
		case CID_CurvePosCsplineBase:  break;
		default:  assert(0);  // currently, no more categories accepted
	}
	
	switch(info->tok_id)
	{
		case TID_pts:  // fall through
		case TID_pts_t:
		{
			bool is_pts_t=(info->tok_id==TID_pts_t);
			NUM::VectorArray<double> pts;
			int rv=_GetVectorArray(pts,info->value,et,is_pts_t ? 4 : 3);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			// Got the point array. 
			// This will remove the data from pts: 
			switch(curvefunc->SetPoints(pts,is_pts_t))
			{
				case 0:  return(ISS_Cont);
				case 1:
					Error(info->value->GetLocationRange(),
						"spline control points were already set\n");
					++n_errors;
					return(ISS_Cont);
				case 2:
					Error(info->value->GetLocationRange(),
						"need at least 2 points for spline\n");
					++n_errors;
					return(ISS_Cont);
				case 3:
					Error(info->value->GetLocationRange(),
						"incompatible \"tvals\" spec\n");
					++n_errors;
					return(ISS_Cont);
				default: assert(0);
			}
		} break;
		case TID_addpts:
		{
			String str;
			int rv=_GetStringValue(str,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(curvefunc->SetAddPoints(str.str()))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				case 2:  return(ISS_IllegalValue);
				default: assert(0);
			}
		} break;
		case TID_tvals:
		{
			String str;
			int rv=_GetStringValue(str,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(curvefunc->SetTValueMode(str.str()))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				case 2:  return(ISS_IllegalValue);
				case 3:
					Error(info->value->GetLocationRange(),
						"tvals spec incompatible with pts_t\n");
					++n_errors;
					return(ISS_Cont);
				default: assert(0);
			}
		} break;
		default: assert(0);
	}
	
	return(ISS_Cont);
}
	

// The tsc MUST be a "function { ... }" scope, i.e. 
// tsc->scope_type->name must be "function". 
int CCObject::_InterpreteCurvePosFunction(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	assert(!"!implemented");
	return(0);
}


// The tsc MUST be a "curve { ... }" scope, i.e. 
// tsc->scope_type->name must be "curve". 
int CCObject::_InterpreteCurvePosCurve(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	assert(!"!implemented");
	return(0);
}


// The tsc MUST be a "join { ... }" scope, i.e. 
// tsc->scope_type->name must be "join". 
int CCObject::_InterpreteCurvePosJoin(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveFunctionBase **_curvefunc)
{
	assert(!"!implemented");
	return(0);
}


int CCObject::_InterpreteCurvePos(InterpreteScopeEntriesInfo *info,
	ExecThread *et,NUM::CurveFunctionBase **_posfunc)
{
	do {
		if(!info->tsc)  break;
		
		int rv=-1;
		switch(info->tsc_tok_id)
		{
			case TID_fixed:   rv=_InterpreteCurvePosFixed(info->tsc,et,_posfunc);   break;
			case TID_lspline: rv=_InterpreteCurvePosLspline(info->tsc,et,_posfunc); break;
			case TID_cspline: rv=_InterpreteCurvePosCspline(info->tsc,et,_posfunc); break;
			case TID_function:rv=_InterpreteCurvePosFunction(info->tsc,et,_posfunc);break;
			case TID_curve:   rv=_InterpreteCurvePosCurve(info->tsc,et,_posfunc);   break;
			case TID_join:    rv=_InterpreteCurvePosJoin(info->tsc,et,_posfunc);    break;
			default: goto breakout;
		}
		
		assert(rv>=0);
		return(rv);
	} while(0); breakout:;
	
	Error(info->value->GetLocationRange(),
		"curve \"pos\" property requires a const/lspline/cspline/"
		"function/curve/join block as value\n");
	++n_errors;
	return(0);
}


int CCObject::_InterpreteCurveTMapNone(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveTMapFunction **_tmapfunc)
{
	return(_InterpreteScope_Template<NUM::CurveTMapFunction_None>(
		"(curve tmap) none",curve_tmap_none_cats,
		/*slot_id=*/CID_CurveTMapNoneBase,
		tsc,et,/*dptr=*/_tmapfunc));
}

int CCObject::_InterpreteCurveTMapNone_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	// Currently no tokens recognized, hence we may not be here. 
	assert(0);
	return(-1);
}


int CCObject::_InterpreteCurveTMapConst(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveTMapFunction **_tmapfunc)
{
	return(_InterpreteScope_Template<NUM::CurveTMapFunction_Linear>(
		"(curve tmap) constant",curve_tmap_const_cats,
		/*slot_id=*/CID_CurveTMapConstBase,
		tsc,et,/*dptr=*/_tmapfunc));
}

int CCObject::_InterpreteCurveTMapConst_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	// Currently no tokens recognized, hence we may not be here. 
	// In the future there will be tokens recognized. 
	assert(0);
	return(-1);
}


int CCObject::_InterpreteCurveTMapCspline(ADTNTypedScope *tsc,ExecThread *et,
	NUM::CurveTMapFunction **_tmapfunc)
{
	assert(!"!implemented");
}


int CCObject::_InterpreteCurveTMap(InterpreteScopeEntriesInfo *info,
	ExecThread *et,NUM::CurveTMapFunction **_tmapfunc,bool is_speed)
{
	do {
		if(!info->tsc)  break;
		
		int rv=-1;
		switch(info->tsc_tok_id)
		{
			case TID_none:
				rv=_InterpreteCurveTMapNone(info->tsc,et,_tmapfunc);    break;
			case TID_constant:
				if(!is_speed)  goto breakout;
				rv=_InterpreteCurveTMapConst(info->tsc,et,_tmapfunc);   break;
			case TID_cspline:
				rv=_InterpreteCurveTMapCspline(info->tsc,et,_tmapfunc); break;
			default:  goto breakout;
		}
		
		assert(rv>=0);
		return(rv);
	} while(0); breakout:;
	
	Error(info->value->GetLocationRange(),
		"curve \"%s\" property requires a none/%scspline"
		"block as value\n",
		is_speed ? "speed" : "length",is_speed ? "constant/" : "");
	++n_errors;
	return(0);
}


// Interprete curve. The tsc MUST be a "curve { ... }" scope, i.e. 
// tsc->scope_type->name must be "curve". 
int CCObject::_InterpreteCurve(ADTNTypedScope *tsc,ExecThread *et,
	NUM::OCurve **_ocurve)
{
	return(_InterpreteScope_Template<NUM::OCurve>(
		"curve",curve_cats,/*slot_id=*/CID_CurveBase,
		tsc,et,/*dptr=*/_ocurve));
}

int CCObject::_InterpreteCurve_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	NUM::OCurve *ocurve=(NUM::OCurve*)info->dptr;
	
	assert(info->cat->cat_id==CID_CurveBase);  // for now until we accept more
	switch(info->tok_id)
	{
		case TID_pos:
		{
			NUM::CurveFunctionBase *posfunc=NULL;
			int rv=_InterpreteCurvePos(info,et,&posfunc);
			if(rv) return(rv);  // never <0
			// Got the curve pos func. 
			switch(ocurve->SetPosCurve(posfunc))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		case TID_dt:
		{
			double dt;
			int rv=_GetScalarValue(dt,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(ocurve->Set_DT(dt))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				case 2:  return(ISS_TimeSpecConflict);
				default: assert(0);
			}
		} break;
		case TID_t:
		{
			Range t;
			int rv=_GetRangeValue(t,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(ocurve->Set_T(t))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				case 2:  return(ISS_TimeSpecConflict);
				default: assert(0);
			}
		} break;
		case TID_speed:
		case TID_length:
		{
			NUM::CurveTMapFunction *tmapfunc=NULL;
			int rv=_InterpreteCurveTMap(info,et,&tmapfunc,
				info->tok_id==TID_speed);
			if(rv) return(rv);  // never <0
			// Got the t mapping func. 
			switch(ocurve->SetTMapFunction(tmapfunc))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				/*case 2:
					Error(info->value->GetLocationRange(),
						"cannot combine \"speed\" and \"length\"\n");
					++n_errors;
					return(ISS_Cont);*/
				default: assert(0);
			}
		} break;
		default: assert(0);
	}
	
	return(ISS_Cont);
}


int CCObject::_InterpreteFrontSpeed(ADTNTypedScope *tsc,ExecThread *et,
	NUM::OFrontSpec **_ofront)
{
	return(_InterpreteScope_Template<NUM::OFrontSpec_Speed>(
		"(front) speed",front_speed_cats,/*slot_id=*/CID_FrontSpeedBase,
		tsc,et,/*dptr=*/_ofront));
}

int CCObject::_InterpreteFrontSpeed_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	// Currently no tokens recognized, hence we may not be here. 
	// In the future there will be tokens recognized. 
	assert(0);
	return(-1);
}


int CCObject::_InterpreteUpGravity(ADTNTypedScope *tsc,ExecThread *et,
	NUM::OUpSpec **_oup)
{
	return(_InterpreteScope_Template<NUM::OUpSpec_Gravity>(
		"(up) gravity",up_gravity_cats,/*slot_id=*/CID_UpGravityBase,
		tsc,et,/*dptr=*/_oup));
}

int CCObject::_InterpreteUpGravity_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	NUM::OUpSpec_Gravity *oup=(NUM::OUpSpec_Gravity*)info->dptr;
	
	assert(info->cat->cat_id==CID_UpGravityBase);
	switch(info->tok_id)
	{
		case TID_force_center:  // fall through
		case TID_force_up:
		{
			bool is_center=(info->tok_id==TID_force_center);
			double tmp[3];
			int rv=_GetVector3Value(tmp,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(oup->SetForce(tmp,is_center))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				case 3:
					Error(info->value->GetLocationRange(),
						"cannot use force_up and force_center simultaniously\n");
					++n_errors;
					return(ISS_Cont);
				default: assert(0);
			}
		} break;
		case TID_accel_scale:
		{
			double tmp;
			int rv=_GetScalarValue(tmp,info->value,et);
			if(rv<0)  {  ++n_errors;  return(0);  }
			if(rv)  return(rv);  // >0 here
			switch(oup->SetAccelScale(tmp))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		default: assert(0);
	}
	
	return(ISS_Cont);
}


int CCObject::_InterpreteMovePos(InterpreteScopeEntriesInfo *info,
	ExecThread *et,NUM::OCurve **_ocurve)
{
	do {
		if(!info->tsc)  break;
		
		int rv=-1;
		switch(info->tsc_tok_id)
		{
			case TID_curve:  rv=_InterpreteCurve(info->tsc,et,_ocurve);  break;
			default:  goto breakout;
		}
		
		assert(rv>=0);
		return(rv);
	} while(0); breakout:;
	
	Error(info->value->GetLocationRange(),
		"move \"pos\" property requires a curve block as value\n");
	++n_errors;
	return(0);
}


int CCObject::_InterpreteMoveFront(InterpreteScopeEntriesInfo *info,
	ExecThread *et,NUM::OFrontSpec **_ofront)
{
	do {
		if(!info->tsc)  break;
		
		int rv=-1;
		switch(info->tsc_tok_id)
		{
			case TID_speed:    rv=_InterpreteFrontSpeed(info->tsc,et,_ofront);  break;
			//case TID_curve:  rv=_InterpreteCurve(info->tsc,et,_ocurve);  break;
			default:  goto breakout;
		}
		
		assert(rv>=0);
		return(rv);
	} while(0); breakout:;
	
	Error(info->value->GetLocationRange(),
		"move \"front\" property requires a curve block as value\n");
	++n_errors;
	return(0);
}	


int CCObject::_InterpreteMoveUp(InterpreteScopeEntriesInfo *info,
	ExecThread *et,NUM::OUpSpec **_oup)
{
	do {
		if(!info->tsc)  break;
		
		int rv=-1;
		switch(info->tsc_tok_id)
		{
			case TID_gravity:  rv=_InterpreteUpGravity(info->tsc,et,_oup);  break;
			//case TID_curve:  rv=_InterpreteCurve(info->tsc,et,_ocurve);  break;
			default:  goto breakout;
		}
		
		assert(rv>=0);
		return(rv);
	} while(0); breakout:;
	
	Error(info->value->GetLocationRange(),
		"move \"front\" property requires a curve block as value\n");
	++n_errors;
	return(0);
}	


int CCObject::_InterpreteMoveScope(ADTNTypedScope *tsc,ExecThread *et,
	NUM::OMovement **_movement)
{
	return(_InterpreteScope_Template<NUM::OMovement>(
		"move ADB",move_cats,/*slot_id=*/CID_MoveBase,
		tsc,et,/*dptr=*/_movement));
}

int CCObject::_InterpreteMoveScope_Slot(ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	NUM::OMovement *movement=(NUM::OMovement*)info->dptr;
	
	assert(info->cat->cat_id==CID_MoveBase);  // for now until we accept more
	switch(info->tok_id)
	{
		case TID_pos:
		{
			NUM::OCurve *ocurve=NULL;
			int rv=_InterpreteMovePos(info,et,&ocurve);
			if(rv)  return(rv);  // only >0 here
			switch(movement->SetPosCurve(ocurve))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		case TID_front:
		{
			NUM::OFrontSpec *ofront=NULL;
			int rv=_InterpreteMoveFront(info,et,&ofront);
			if(rv)  return(rv);  // only >0 here
			switch(movement->SetFrontSpec(ofront))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		case TID_up:
		{
			NUM::OUpSpec *oup=NULL;
			int rv=_InterpreteMoveUp(info,et,&oup);
			if(rv)  return(rv);  // only >0 here
			switch(movement->SetUpSpec(oup))
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);
				default: assert(0);
			}
		} break;
		case TID_front_mode:  // fall through
		case TID_up_mode:
		{
			String str;
			int rv=_GetStringValue(str,info->value,et);
			if(rv<0)  {  ++n_errors;  return(ISS_Cont);  }
			if(rv)  return(rv);  // >0 here
			switch(info->tok_id==TID_front_mode ? 
				movement->SetFrontMode(str.str()) : 
				movement->SetUpMode(str.str()) )
			{
				case 0:  return(ISS_Cont);
				case 1:  return(ISS_AlreadySet);  // <-- Will not happen: new settings override old ones. 
				case 2:  return(ISS_IllegalValue);
				default: assert(0);
			}
		} break;
		default: assert(0);
	}
	
	return(ISS_Cont);
}


int CCObject::InterpreteScopeSlot(int slot_id,ExecThread *et,
	InterpreteScopeEntriesInfo *info)
{
	fprintf(stderr,"  slot[%d] -> \"%s\" (%d/%d)\n",
		slot_id,AniDescBlockTokIDString(info->tok_id),
		info->cat->cat_id,info->tok_id);
	switch(slot_id)
	{
		case CID_MoveBase:               return(_InterpreteMoveScope_Slot(et,info));
		case CID_CurveBase:              return(_InterpreteCurve_Slot(et,info));
		case CID_CurvePosFixedBase:      return(_InterpreteCurvePosFixed_Slot(et,info));
		case CID_CurvePosLsplineBase:    // fall through 
		case CID_CurvePosCsplineBase:    return(_InterpreteCurvePosLCspline_Slot(et,info));
		case CID_CurveTMapNoneBase:      return(_InterpreteCurveTMapNone_Slot(et,info));
		case CID_FrontSpeedBase:         return(_InterpreteFrontSpeed_Slot(et,info));
		case CID_UpGravityBase:          return(_InterpreteUpGravity_Slot(et,info));
		default: assert(0);
	}
	return((InterpreteScopeSlot_State)-1);  // not reached
}


// Exclusively used by EvalAniDescBlock(). 
int CCObject::_DoEvalAniDescBlock(ADTNAniDescBlock *adb,ExecThread *et)
{
	for(ANI::TreeNode *_tn=et->ess->popTN(adb->down.first()); 
		_tn; _tn=_tn->next)
	{
		if(_tn->NType()!=ANI::TN_AniDesc) assert(0);
		if(((ADTreeNode*)_tn)->ADNType()!=ADTN_ScopeEntry) assert(0);
		
		ADTNScopeEntry *scent=(ADTNScopeEntry*)_tn;
		ANI::TNIdentifier *name = (scent->down.first()==scent->down.last() ? 
			NULL : (ANI::TNIdentifier*)scent->down.first());
		ANI::TreeNode *value=scent->down.last();
		
		if(name)
		{
			Error(name->GetLocationRange(),
				"unknown property \"%s\" in object toplevel ADB [ignored]\n",
				name->CompleteStr().str());
			++n_errors;
			continue;
		}
		
		if(value->NType()==ANI::TN_AniDesc && 
			((ADTreeNode*)value)->ADNType()==ADTN_TypedScope)
		{
			ADTNTypedScope *tsc=(ADTNTypedScope*)value;
			
			RefString tsc_type(tsc->scope_type->name);
			if(tsc_type=="move")
			{
				NUM::OMovement *movement=NULL;
				if(_InterpreteMoveScope(tsc,et,&movement))
				{
					// Context switch. 
					et->ess->pushTN(_tn);
					return(1);
				}
				
				// Got a new movement; add to list. 
				movements.append(movement);
				
				// Finally create the movement: 
				// This will generate all the internal data and also 
				// try to achieve continuacy,... 
				int rv=movement->Create(setting->CurrentSettingTime());
				if(rv)
				{  ++n_errors;  }
				
				// FIXME: Need to check if successive movements overlap 
				// in time (using movement->{T0,T1}(). 
			}
			else
			{
				Error(tsc->GetLocationRange(),
					"cannot use \"%s\" scope in toplevel object ADB [ignored]\n",
					tsc_type.str());
				++n_errors;
				continue;
			}
		}
		else
		{
			Error(value->GetLocationRange(),
				"illegal (non-scope) subnode in toplevel object ADB [ignored]\n");
			++n_errors;
			continue;
		}
	}
	
	return(0);
}


int CCObject::EvalAniDescBlock(ADTNAniDescBlock *adb,ExecThread *et)
{
	// We must make sure that only one thread at a time executes an 
	// ani desc block. 
	if(AquireADBLock(adb,et))
	{  return(0);  }  // <-- No context switch (could use 1 for waiting, too). 
	
	fprintf(stdout,"CC: Interpreting ADB @%s\n",
		adb->GetLocationRange().PosRangeString().str());
	
	int rv=_DoEvalAniDescBlock(adb,et);
	
	if(rv)
	{
		// Context switch. 
		if(et->schedule_status==ExecThread::SSRequest)
		{
			Error(adb->GetLocationRange(),
				"OOPS: Cloning from inside ADB forbidden\n");
			abort();
		}
		
		return(1);
	}
	
	// If we come here, there is regular exit, i.e. no context switch. 
	LoseADBLock();  // Does NOT clear n_errors. 
	return(n_errors ? -1 : 0);
}


NUM::OMovement *CCObject::_FindMovementByTime(double setting_time)
{
	// First, query cache: 
	do {
		if(!_prev_looked_up_movement)  break;
		if(_prev_looked_up_movement->T0()<=setting_time && 
			_prev_looked_up_movement->T1()>setting_time)
		{  return(_prev_looked_up_movement);  }
		
		// See if it happens to be the next movement: 
		NUM::OMovement *next=_prev_looked_up_movement->next;
		if(!next)  break;
		if(next->T0()<=setting_time && next->T1()>setting_time)
		{  _prev_looked_up_movement=next;  return(next);  }
	} while(0);
	
	// Perform lookup: 
	NUM::OMovement *close=NULL;  // Just the > vs >= issue...
	for(NUM::OMovement *m=movements.first(); m; m=m->next)
	{
		if(m->T0()<=setting_time && m->T1()>=setting_time)
		{
			close=m;
			if(m->T1()>setting_time)  break;
		}
	}
	_prev_looked_up_movement=close;
	return(close);
}


int CCObject::GetVal(OVal ov,int *store_here)
{
	fprintf(stderr,"CC:S: Implement GetVal()\n");
	return(CCIF_Object::GetVal(ov,store_here));
}


int CCObject::GetVal(OVal ov,double *store_here)
{
	// First, get the current movement: 
	double setting_time = setting->CurrentSettingTime();
	NUM::OMovement *m=_FindMovementByTime(setting_time);
	
	switch(ov)
	{
		case ValTime:
			// FIXME: Should return object time(?) For now, I am using setting time. 
			// Only used in povwriter to write a comment, currently. 
			*store_here=setting_time;
			return(0);
		case ValPos:
			if(!m)
			{  NUM::OMovement::DefaultPos(store_here);  return(0);  }
			m->GetPos(setting_time,store_here);
			return(0);
		case ValFront:
			if(!m)
			{  NUM::OMovement::DefaultFront(store_here);  return(0);  }
			m->GetFront(setting_time,store_here,/*pos_already_known=*/NULL);
			return(0);
		case ValUp:
			if(!m)
			{  NUM::OMovement::DefaultUp(store_here);  return(0);  }
			m->GetUp(setting_time,store_here,
				/*pos_already_known=*/NULL,/*front_already_known=*/NULL);
			return(0);
		case ValSpeed:
		case ValAccel:
			assert(!"!implemented");
			return(0);
		// NO: default: assert(0); NO!
	}
	
	// Return correct error value: 
	return(CCIF_Object::GetVal(ov,store_here));
}


int CCObject::GetVal(OVal ov,NUM::ObjectMatrix *store_here)
{
	if(ov==ValMatrix)
	{
		// First, get the current movement: 
		double setting_time = setting->CurrentSettingTime();
		NUM::OMovement *m=_FindMovementByTime(setting_time);
		if(m)
		{
			double pos[3],front[3],up[3];

			m->GetPos(setting_time,pos);
			m->GetFront(setting_time,front,
				/*pos_already_known=*/pos);
			m->GetUp(setting_time,up,
				/*pos_already_known=*/pos,
				/*front_already_known=*/front);

			// FIXME: (a) allow right vector; (b) use obj_{front,right,up}
			store_here->SetObjPos(pos,front,up,NULL);
			return(0);
		}
		else
		{
			// No movement for current time. 
			assert(!"!implemented");
		}
	}
	
	// *store_here=NUM::ObjectMatrix::IdentMat;
	
	// Return correct error value: 
	return(CCIF_Object::GetVal(ov,store_here));
}


CCObject::CCObject(CCSetting *_setting,int _vflags) : 
	CCIF_Object(_setting,_vflags),
	LinkedListBase<CCObject>(),
	movements()
{
	Setting()->RegisterObject(this);
}

CCObject::~CCObject()
{
	Setting()->UnregisterObject(this);
	
	while(!movements.is_empty())
	{  delete movements.popfirst();  }
}

}}  // end of namespace CC::S
