/*
 * core/a_ivarobj.cc
 * 
 * ANI internal variables which are member of animation objects. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "animation.h"
#include "execthread.h"
#include "aniinstance.h"

using namespace ANI;


static void _DoRegisterVar(AniAniSetObj *parent,
	AniInternalVariable::BuiltinVarType bvt,const char *name)
{
	RefString tmp;
	tmp.set(name);
	new AniInternalVariable_Object(tmp,bvt,NULL,parent);
}


void AniInternalVariable_Object::_RegisterInternalVariables(
	AniAniSetObj *parent)
{
	if(parent->ASType()==AST_Object)
	{
		_DoRegisterVar(parent,BVT_Time,"$time");
		_DoRegisterVar(parent,BVT_Pos,"$pos");
		_DoRegisterVar(parent,BVT_Front,"$front");
		_DoRegisterVar(parent,BVT_Up,"$up");
		_DoRegisterVar(parent,BVT_Speed,"$speed");
		_DoRegisterVar(parent,BVT_Accel,"$accel");
	}
}

//------------------------------------------------------------------------------

ExprValue AniInternalVariable_Object::EAF_Member::Eval(TNIdentifier *idf,
	ExecThreadInfo *info)
{
	const AniGlue_ScopeInstanceBase *asic=info->stack->GetThisPtr().GetASIB();
	// If this assert fails, we accepted "this" in a location where 
	// no "this" pointer is defined which is an error in the TF step. 
	// OR, the "this" pointer is NULL due to user spec which is 
	// a runtime error here. 
	assert(asic);
	ExprValue tmp;
	int rv=((AniScopeInstanceBase*)asic)->GetInternalVar(bvt,tmp);
	// If this assert fails, we had an error in GetInternalVar(). 
	// Should be handeled by prompting the user and aborting, probably. (see also below)
	assert(rv==0);
	return(tmp);
}

ExprValue AniInternalVariable_Object::EAF_Member::Eval(TNIdentifier *idf,
	const ExprValue &base)
{
	ScopeInstance inst;
	base.GetAniScope(&inst);
	AniGlue_ScopeInstanceBase *asic=inst.GetASIB();   // get internal instance base
	if(!asic)  _NullScopeExplicitMemberSelect(idf);
	ExprValue tmp;
	int rv=((AniScopeInstanceBase*)asic)->GetInternalVar(bvt,tmp);
	// If this assert fails, we had an error in GetInternalVar(). 
	// Should be handeled by prompting the user and aborting, probably. (see also above)
	assert(rv==0);
	return(tmp);
}

ExprValue &AniInternalVariable_Object::EAF_Member::GetRef(TNIdentifier * /*idf*/,
	ExecThreadInfo * /*info*/)
{
	// Currently, all internal object vars are no lvalues. 
	assert(0);
}

AniInternalVariable_Object::EAF_Member::~EAF_Member()
{
	// Not inline because virtual. 
}


//------------------------------------------------------------------------------


ExprValueType AniInternalVariable_Object::GetType() const
{
	switch(bvt)  // Fall through switch. 
	{
		case BVT_Time:
		{
			Value::CompleteType tmp;
			tmp.type=Value::VTScalar;
			return(ExprValueType(tmp));
		}
		case BVT_Pos:
		case BVT_Front:
		case BVT_Up:
		case BVT_Speed:
		case BVT_Accel:
		{
			Value::CompleteType tmp;
			tmp.type=Value::VTVector;
			tmp.n=3;
			return(ExprValueType(tmp));
		}
		default: assert(0);
	}
	return ExprValueType();  // not reached
}


int AniInternalVariable_Object::CG_GetVariableType(VariableTypeInfo *vti) const
{
	vti->store_here->operator=(GetType());
	vti->is_lvalue=0;
	vti->is_const=0;
	vti->may_ieval=0;
	return(0);
}


AniGlue_ScopeBase::EvalAddressationFunc *AniInternalVariable_Object::CG_GetVarEvalAdrFuncFunc(
	TNIdentifier *idf_for_debug)
{
	return(GetVarEvalAdrFuncFunc(idf_for_debug,/*ainc=*/NULL));
}


AniGlue_ScopeBase::EvalAddressationFunc *AniInternalVariable_Object::GetVarEvalAdrFuncFunc(
	TNIdentifier * /*idf_for_debug*/,AniIncomplete * /*ainc*/)
{
	// We're (non-static) member of an object. Hence: 
	assert(Parent()->ASType()==AST_Object);
	return(new AniInternalVariable_Object::EAF_Member(bvt));
}


void AniInternalVariable_Object::AniTreeDump(StringTreeDump *d)
{
	d->Append(AniScopeTypeStr());
	d->Append(" ");
	d->Append(GetType().TypeString().str());
	d->Append(" ");
	d->Append(name.str());
	d->Append(":");
	d->Append(" (member)");
	d->Append("\n");
}


AniInternalVariable_Object::AniInternalVariable_Object(
	const RefString &_name,BuiltinVarType _bvt,
	ANI::TreeNode *_treenode,AniScopeBase *_parent) : 
	AniInternalVariable(_name,_treenode,_parent),bvt(_bvt)
{
}

AniInternalVariable_Object::~AniInternalVariable_Object()
{
}
