/*
 * core/i_object.cc
 * 
 * Animation object instanc. 
 * This is part of the AniVision project. 
 * 
 * Copyright (c) 2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "aniinstance.h"
#include "animation.h"
#include <calccore/ccif.h>

using namespace ANI;


int AniObjectInstance::GetInternalVar(BuiltinVarType bvt,
	ExprValue &store_val)
{
	char vtype='\0';  // 'd'ouble, '3'd vector, 'i'nteger
	CC::CCIF_Object::OVal oval;
	switch(bvt)
	{
		#define AV AniInternalVariable
		case AV::BVT_Time:   vtype='d'; oval=CC::CCIF_Object::ValTime;   break;
		case AV::BVT_Pos:    vtype='3'; oval=CC::CCIF_Object::ValPos;    break;
		case AV::BVT_Front:  vtype='3'; oval=CC::CCIF_Object::ValFront;  break;
		case AV::BVT_Up:     vtype='3'; oval=CC::CCIF_Object::ValUp;     break;
		case AV::BVT_Speed:  vtype='3'; oval=CC::CCIF_Object::ValSpeed;  break;
		case AV::BVT_Accel:  vtype='3'; oval=CC::CCIF_Object::ValAccel;  break;
		case AV::BVT_Active: vtype='i'; oval=CC::CCIF_Object::ValActive; break;
		#undef AV
		default: assert(0);
	}
	
	int rv=-2;
	do {
		if(!ccif)  break;  // rv stays -2
		switch(vtype)
		{
			case 'i':
			{
				int tmp;
				rv=ccif->GetVal(oval,&tmp);
				if(rv) break;
				store_val.assign(tmp);
			}	break;
			case 'd':
			{
				double tmp;
				rv=ccif->GetVal(oval,&tmp);
				if(rv) break;
				store_val.assign(tmp);
			}	break;
			case '3':
			{
				// This is a little hack...
				Vector tmp(Vector::NoInit,3);
				rv=ccif->GetVal(oval,&tmp[0]);  // tmp[3] returns ref. 
				if(rv) break;
				store_val.assign(tmp);
			}	break;
			default: assert(0);
		}
	} while(0);
	
	// Store null values in case of error. 
	if(rv) switch(vtype)
	{
		case 'i':  store_val.assign((int)0);  break;
		case 'd':  store_val.assign((double)0.0);  break;
		case '3':  store_val.assign(Vector(Vector::Null,3));  break;
		default: assert(0);
	}
	
	return(rv);
}


AniAnimation *AniObjectInstance::_GetAnimation()
{
	#warning "Could be replaced by O(1) implementation."
	for(AniScopeBase *i=asb; i; i=i->Parent())
	{
		if(i->ASType()==AniScopeBase::AST_Animation)
		{  return((AniAnimation*)i);  }
	}
	assert(0);  // Must have animation in hierarchy. 
	return(NULL);
}


TNCompoundStmt *AniObjectInstance::_GetPOVCompound()
{
	return( ((TNFunctionDef*)(
			((AniObject*)asb)->pov_attach_detach->GetTreeNode()
		))->func_body );
}


AniObjectInstance::AniObjectInstance(AniScopeBase *_asb) : 
	AniScopeInstanceBase(_asb),
	del_notifier_list()
{
	// We may only create an object if we have an animation 
	// context. 
	// (Exception: objects which do not need internal vars could 
	// come along without ever attaching a CCIF and could also be 
	// created in static init and inside animation block.) 
	//   THIS IS NOT YET IMPLEMENTED as we assume that every object 
	//   needs animation values. 
	
	// Check for presence of animation context: 
	AniAnimation *ani=_GetAnimation();
	AniAnimation::Context *active_context=ani->GetActiveContext();
	if(!active_context)
	{
		// Woulc be nice if we could report the _ERROR_ location here. 
		Error(asb->NameLocation(),
			"attempt to create object of type %s while not in animation "
			"context\n",
			asb->CompleteName().str());
		abort();
	}
	
	// Okay, have to attach the calc core interface: 
	ccif=ani->Get_CCIF_Factory()->CreateCCIF_Object(active_context->ccif,
		/*vflags=*/CC::CCIF_Object::VF_TimeLocAct);
}

AniObjectInstance::~AniObjectInstance()
{
	// Call the delete notifiers: 
	// These notifiers MUST DELETE ITSELF, so this works: 
	while(!del_notifier_list.is_empty())
	{  del_notifier_list.first()->DeleteNotify(this);  }
	
	DELETE(ccif);
}

